%token <Big_int.big_int> INT
%token <float> FLOAT
%token <string> ATOM
%token <string> STRING
%token <string> FUNCTOR
%token <string> IDENT IDENT_PAR IDENT_SQ
%token <Kernel.Const.t> KERNEL_CONST
%token XML_TAG_START XML_TAG_END XML_ELEMENT_END
%token COMMA DOT DOTDOT DOTDOTDOT COLON SEMICOLON EOF
%token PAR_OPEN PAR_CLOSE PAR_SQ SQ_OPEN SQ_CLOSE SQ_SQ CURLY_OPEN CURLY_CLOSE
%token ARROW DBL_ARROW LEFT_ARROW
%token EQ NEQ LT LE GT GE COLON_EQ
%token PLUS MINUS MUL DIV POW
%token UNDERSCORE AND CALL DO ELSE EXISTS FOR FORALL FUNCTION IF IN LET MOD
%token MODULE NAMESPACE NOT OR PACKAGE THIS WITH
%type <string list option * Ast.t> main
%start main
%nonassoc DOTDOTDOT
%right SEMICOLON
%right COMMA
%nonassoc ARROW DBL_ARROW
%nonassoc NOT
%nonassoc LEFT_ARROW
%right OR
%right AND
%nonassoc EQ NEQ LT LE GT GE COLON_EQ
%nonassoc DOTDOT
%left PLUS MINUS
%left MUL DIV MOD
%right POW
%left DOT COLON
%nonassoc DO
%%

main:
| p = package? e = expr EOF { (p, e) }
| error {
    Err.syntax {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    }
  }

package:
| PACKAGE p = separated_nonempty_list(DOT, IDENT) { p }

expr:
| l = expr DOTDOTDOT {
    { Pos.v = Ast.Dotdotdot l;
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = expr SEMICOLON r = expr {
    { Pos.v = Ast.Semicolon (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = expr COMMA r = expr {
    { Pos.v = Ast.Comma (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = expr ARROW r = expr {
    { Pos.v = Ast.Arrow (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = expr DBL_ARROW r = expr {
    { Pos.v = Ast.Dbl_arrow (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| MINUS e = simple_expr {
    { Pos.v = 
        Ast.Uminus 
          { Pos.v = e;
            range = {
              Pos.first = Pos.of_lexing $startpos;
              Pos.last = Pos.of_lexing $endpos
            };
          };
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| e = concat_expr { e }

concat_expr:
| {
    { Pos.v = Ast.Empty;
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| e = non_empty_concat_expr { e }

non_empty_concat_expr:
| l = sub_expr r = concat_expr {
    { Pos.v = Ast.Concat (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| FORALL v = IDENT+ COMMA e = concat_expr {
    { Pos.v = Ast.Forall (v, e);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| FORALL v = IDENT+ l = delimited_expr r = concat_expr {
    let range = {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    } in
    { Pos.v = Ast.Concat ({ Pos.v = Ast.Forall (v, l); range = range }, r);
      range = range;
    }
  }
| FORALL v = IDENT* w = IDENT_PAR l = expr PAR_CLOSE r = concat_expr {
    let range = {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    } in
    let l = { Pos.v = Ast.Par l; range = range } in
    { Pos.v = Ast.Concat ({ Pos.v = Ast.Forall (v @ [w], l); range = range }, r);
      range = range;
    }
  }
| EXISTS v = IDENT+ COMMA r = concat_expr {
    let range = {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    } in
    { Pos.v = Ast.Concat ({ Pos.v = Ast.Exists_comma v; range = range }, r);
      range = range
    }
  }
| EXISTS v = IDENT+ l = delimited_expr r = concat_expr {
    let range = {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    } in
    { Pos.v = Ast.Concat ({ Pos.v = Ast.Exists (v, l); range = range }, r);
      range = range;
    }
  }
| EXISTS v = IDENT* w = IDENT_PAR l = expr PAR_CLOSE r = concat_expr {
    let range = {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    } in
    let l = { Pos.v = Ast.Par l; range = range } in
    { Pos.v = Ast.Concat ({ Pos.v = Ast.Exists (v @ [w], l); range = range }, r);
      range = range;
    }
  }
| FOR x = let_patt IN a = sub_expr b = sub_expr r = concat_expr {
    let range = {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    } in
    { Pos.v = Ast.Concat ({ Pos.v = Ast.For (x, a, b); range = range }, r);
      range = range;
    }
  }
| n = namespace r = concat_expr {
    { Pos.v = Ast.Concat (n, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| f = fct r = concat_expr {
    { Pos.v = Ast.Concat (f, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }

namespace:
| NAMESPACE p = separated_nonempty_list(DOT, IDENT) {
    { Pos.v = Ast.Namespace p;
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }

fct:
| FUNCTION e = sub_expr {
    { Pos.v = Ast.Function e;
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }

sub_expr:
| l = sub_expr LEFT_ARROW r = sub_expr {
    { Pos.v = Ast.Left_arrow (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| DO e = sub_expr { 
    let range = {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    } in
    { Pos.v = Ast.Do e;
      range = range;
    }
  }
| LET v = let_patt EQ e = sub_expr { 
    let range = {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    } in
    { Pos.v = Ast.Let (v, e);
      range = range
    }
  }
| l = sub_expr OR r = sub_expr {
    { Pos.v = Ast.Or (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr AND r = sub_expr {
    { Pos.v = Ast.And (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| NOT l = sub_expr {
    { Pos.v = Ast.Not l;
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr EQ r = sub_expr {
    { Pos.v = Ast.Eq (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr COLON_EQ r = sub_expr {
    { Pos.v = Ast.Colon_eq (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr NEQ r = sub_expr {
    { Pos.v = Ast.Neq (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr LT r = sub_expr {
    { Pos.v = Ast.Lt (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr LE r = sub_expr {
    { Pos.v = Ast.Le (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr GT r = sub_expr {
    { Pos.v = Ast.Gt (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr GE r = sub_expr {
    { Pos.v = Ast.Ge (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr DOT r = sub_expr {
    let range = {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    } in
    match r.Pos.v with
      | Ast.Get (l', r') -> {
          Pos.v = Ast.Get ({ Pos.v = Ast.Dot (l, l'); range = range }, r');
          range = range
        }
      | _ -> { Pos.v = Ast.Dot (l, r); range = range }
  }
| l = sub_expr DOTDOT r = sub_expr {
    { Pos.v = Ast.Dotdot (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr COLON r = sub_expr {
    { Pos.v = Ast.Colon (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr PLUS r = sub_expr {
    { Pos.v = Ast.Plus (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr MINUS r = sub_expr {
    { Pos.v = Ast.Minus (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr MUL r = sub_expr {
    { Pos.v = Ast.Mul (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr DIV r = sub_expr {
    { Pos.v = Ast.Div (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr MOD r = sub_expr {
    { Pos.v = Ast.Mod (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| l = sub_expr POW r = sub_expr {
    { Pos.v = Ast.Pow (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| a = simple_expr {
    { Pos.v = a;
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| a = delimited_expr { a }

par_cont:
| PAR_CLOSE { fun i -> i }
| PAR_SQ args = expr k = sq_cont {
    fun f ->
      let f =
        { Pos.v = f;
          range = {
            Pos.first = Pos.of_lexing $startpos;
            Pos.last = Pos.of_lexing $endpos
          }
        } in
      k (Ast.Get (f, args))
  }

sq_cont:
| SQ_CLOSE { fun i -> i }
| SQ_SQ args = expr k = sq_cont {
    fun f ->
      let f =
        { Pos.v = f;
          range = {
            Pos.first = Pos.of_lexing $startpos;
            Pos.last = Pos.of_lexing $endpos
          }
        } in
      k (Ast.Get (f, args))
  }

let_patt:
| e = simple_expr {
    { Pos.v = e;
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| e = delimited_expr { e }
| l = let_patt COMMA r = let_patt {
    { Pos.v = Ast.Comma (l, r);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }

simple_expr:
| UNDERSCORE { Ast.Underscore }
| THIS { Ast.This }
| c = KERNEL_CONST { Ast.Kernel_const c }
| s = ATOM { Ast.Atom s }
| i = INT { Ast.Const (Kernel.Const.Int i) }
| f = FLOAT { Ast.Const (Kernel.Const.Float f) }
| s = STRING { Ast.Const (Kernel.Const.String s) }
| v = IDENT { Ast.Var v }
| f = IDENT_PAR args = expr k = par_cont {
    k (Ast.Tok (f, args))
  }
| v = IDENT_SQ args = expr k = sq_cont {
    k (Ast.Get ({ Pos.v = Ast.Var v; 
          range = {
            Pos.first = Pos.of_lexing $startpos;
            Pos.last = Pos.of_lexing $endpos
          }}, args))
  }
| f = FUNCTOR args = expr k = par_cont {
    k (Ast.Functor (f, args))
  }
| CALL f = IDENT_PAR args = expr PAR_CLOSE
    { Ast.Call (f, args) }
| MODULE e = delimited_expr { Ast.Module e }
| SQ_OPEN items = expr SQ_CLOSE {
    Ast.List items
  }
| XML_TAG_START e = expr XML_ELEMENT_END {
    let empty = { Pos.v = Ast.Empty;
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    } in
    Ast.Xml (e, empty, empty)
  }
| XML_TAG_START e0 = expr XML_TAG_END e1 = expr
  XML_ELEMENT_END e2 = expr XML_TAG_END { Ast.Xml (e0, e1, e2) }
| e = conditional { e }

conditional:
| WITH e = sub_expr t = conditional_expr { Ast.With (e, t) }
| IF v = if_vars e = sub_expr t = conditional_expr { Ast.If (v, e, t) }
| IF v = if_vars e = sub_expr t = conditional_expr ELSE f = conditional_else
    { Ast.IfElse (v, e, t, f )} 

if_vars:
| { [] }
| EXISTS v = IDENT+ COMMA { v }

conditional_else:
| c = conditional {
    { Pos.v = c;
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| e = conditional_expr { e }

conditional_expr:
| e = delimited_expr { e }
| DO e = delimited_expr { 
    let range = {
      Pos.first = Pos.of_lexing $startpos;
      Pos.last = Pos.of_lexing $endpos
    } in
    { Pos.v = Ast.Do e;
      range = range;
    }
  }

delimited_expr:
| CURLY_OPEN a = expr CURLY_CLOSE {
    { Pos.v = Ast.Curly a;
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
| PAR_OPEN a = expr k = par_cont {
    { Pos.v = k (Ast.Par a);
      range = {
        Pos.first = Pos.of_lexing $startpos;
        Pos.last = Pos.of_lexing $endpos
      }
    }
  }
